/*################################################################################
  ##
  ##   Copyright (C) 2016-2020 Keith O'Hara
  ##
  ##   This file is part of the OptimLib C++ library.
  ##
  ##   Licensed under the Apache License, Version 2.0 (the "License");
  ##   you may not use this file except in compliance with the License.
  ##   You may obtain a copy of the License at
  ##
  ##       http://www.apache.org/licenses/LICENSE-2.0
  ##
  ##   Unless required by applicable law or agreed to in writing, software
  ##   distributed under the License is distributed on an "AS IS" BASIS,
  ##   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  ##   See the License for the specific language governing permissions and
  ##   limitations under the License.
  ##
  ################################################################################*/

#include "matrix_ops/abs.hpp"
#include "matrix_ops/abs_max.hpp"
#include "matrix_ops/accu.hpp"
#include "matrix_ops/access.hpp"
#include "matrix_ops/array_add.hpp"
#include "matrix_ops/array_div.hpp"
#include "matrix_ops/as_scalar.hpp"
#include "matrix_ops/cos.hpp"
#include "matrix_ops/cout.hpp"
#include "matrix_ops/diagmat.hpp"
#include "matrix_ops/diagvec.hpp"
#include "matrix_ops/dot.hpp"
#include "matrix_ops/endl.hpp"
#include "matrix_ops/eval.hpp"
#include "matrix_ops/exp.hpp"
#include "matrix_ops/eye.hpp"
#include "matrix_ops/hadamard.hpp"
#include "matrix_ops/is_finite.hpp"
#include "matrix_ops/inv.hpp"
#include "matrix_ops/log.hpp"
#include "matrix_ops/max.hpp"
#include "matrix_ops/min.hpp"
#include "matrix_ops/ncol.hpp"
#include "matrix_ops/norm.hpp"
#include "matrix_ops/ones.hpp"
#include "matrix_ops/pow.hpp"
#include "matrix_ops/randi.hpp"
#include "matrix_ops/randn.hpp"
#include "matrix_ops/randu.hpp"
#include "matrix_ops/set_size.hpp"
#include "matrix_ops/size.hpp"
#include "matrix_ops/solve.hpp"
#include "matrix_ops/sqrt.hpp"
#include "matrix_ops/sum.hpp"
#include "matrix_ops/transpose.hpp"
#include "matrix_ops/zeros.hpp"

//

#include "matrix_ops/get_sort_index.hpp"
#include "matrix_ops/index_min.hpp"
#include "matrix_ops/reset_negative_values.hpp"
